package com.perforce.maven.mojo;

import java.io.File;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugins.annotations.Component;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.MavenProject;
import org.apache.maven.scm.ScmException;
import org.apache.maven.scm.log.ScmLogDispatcher;
import org.apache.maven.scm.manager.ScmManager;
import org.apache.maven.scm.provider.ScmProviderRepositoryWithHost;
import org.apache.maven.scm.repository.ScmRepository;
import org.apache.maven.settings.Server;
import org.apache.maven.settings.Settings;
import org.codehaus.plexus.util.StringUtils;
import org.sonatype.plexus.components.sec.dispatcher.SecDispatcher;
import org.sonatype.plexus.components.sec.dispatcher.SecDispatcherException;

import com.perforce.maven.scm.provider.p4.manager.P4ClientManager;
import com.perforce.p4java.core.IChangelist;

/**
 * Super class for Perforce Mojos.
 * 
 * @requiresProject
 */
public abstract class P4Mojo
    extends AbstractMojo
{

    /** The Constant P4_CHANGELIST_PROPERTY. */
    public static final String P4_CHANGELIST_PROPERTY = "maven.scm.p4.changelist";

    /** The Constant P4_CHANGELIST_PROPERTY. */
    public static final String P4_CHANGELIST_DESCRIPTION_PROPERTY = "maven.scm.p4.changelist.description";

    /** The Constant P4_FROM_CHANGELIST_PROPERTY. */
    public static final String P4_FROM_CHANGELIST_PROPERTY = "maven.scm.p4.fromchangelist";

    /** The Constant P4_TO_CHANGELIST_PROPERTY. */
    public static final String P4_TO_CHANGELIST_PROPERTY = "maven.scm.p4.tochangelist";

    /**
     * @parameter expression="${project.scm.developerConnection}"
     * @readonly
     */
    @Parameter( defaultValue = "${project.scm.developerConnection}" )
    protected String urlScm;

    /**
     * The username that is used when connecting to the SCM system.
     */
    @Parameter( property = "username" )
    protected String username;

    /**
     * The password that is used when connecting to the SCM system.
     */
    @Parameter( property = "password" )
    protected String password;

    /**
     * The includes that is used to include files with that pattern.
     */
    @Parameter( property = "includes" )
    protected String includes;

    /**
     * The excludes that is used to exclude files with that pattern.
     */
    @Parameter( property = "excludes" )
    protected String excludes;

    /**
     * Local directory to be used to issue SCM actions
     */
    @Parameter( defaultValue = "${project.basedir}" )
    protected File scmDirectory;

    @Component
    protected ScmManager scmManager;

    /**
     * The maven project.
     */
    @Parameter( defaultValue = "${project}", readonly = true )
    protected MavenProject project;

    /**
     * When this plugin requires Maven 3.0 as minimum, this component can be removed and o.a.m.s.c.SettingsDecrypter be
     * used instead.
     */
    @Component( hint = "mng-4384" )
    private SecDispatcher secDispatcher;

    @Parameter( defaultValue = "${settings}", readonly = true )
    private Settings settings;

    @Component
    protected P4ClientManager p4ClientManager;

    protected ScmLogDispatcher logger;

    protected ScmRepository getScmRepository()
        throws ScmException
    {
        ScmRepository repository = null;

        if ( StringUtils.isEmpty( urlScm ) )
        {
            getLog().error( "You need to define a developerConnectionUrl parameter." );
            return null;
        }

        try
        {
            repository = scmManager.makeScmRepository( urlScm );

            ScmProviderRepositoryWithHost scmRepo = (ScmProviderRepositoryWithHost) repository.getProviderRepository();

            loadInfosFromSettings( scmRepo );

            if ( !StringUtils.isEmpty( username ) )
            {
                scmRepo.setUser( username );
            }

            if ( !StringUtils.isEmpty( password ) )
            {
                scmRepo.setPassword( password );
            }
        }
        catch ( Exception e )
        {
            getLog().error( "Problem connecting to the repository with the specified developerConnectionUrl parameter.",
                            e );
        }

        return repository;
    }

    /**
     * Parse the changelist string to a changelist number. Convert the "default" changelist string to the default
     * changelist number. If it is negative return unknown changelist. Otherwise, return the converted changelist
     * number.
     * 
     * @param changelist the changelist
     * @return the int
     */
    protected static int parseChangelist( String changelist )
    {
        if ( StringUtils.isNotEmpty( changelist ) )
        {
            if ( changelist.trim().equalsIgnoreCase( "default" ) )
            {
                return IChangelist.DEFAULT;
            }
            try
            {
                int changelistId = Integer.parseInt( changelist );
                if ( changelistId < 0 )
                {
                    return IChangelist.UNKNOWN;
                }
                return changelistId;
            }
            catch ( NumberFormatException e )
            {
                // Suppress error
            }
        }
        return IChangelist.UNKNOWN;
    }

    private String decrypt( String str, String server )
    {
        try
        {
            return secDispatcher.decrypt( str );
        }
        catch ( SecDispatcherException e )
        {
            getLog().warn( "Failed to decrypt password/passphrase for server " + server + ", using auth token as is" );
            return str;
        }
    }

    /**
     * Load username password from settings if user has not set them in JVM properties
     * 
     * @param repo not null
     */
    private void loadInfosFromSettings( ScmProviderRepositoryWithHost repo )
    {
        if ( username == null || password == null )
        {
            String host = repo.getHost();

            int port = repo.getPort();

            if ( port > 0 )
            {
                host += ":" + port;
            }

            Server server = this.settings.getServer( host );

            if ( server != null )
            {
                if ( username == null )
                {
                    username = server.getUsername();
                }

                if ( password == null )
                {
                    password = decrypt( server.getPassword(), host );
                }
            }
        }
    }
}
