package com.perforce.maven.scm.provider.p4.manager;

import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.io.FileUtils;
import org.apache.maven.scm.ScmException;
import org.codehaus.plexus.logging.AbstractLogEnabled;
import org.codehaus.plexus.personality.plexus.lifecycle.phase.Disposable;
import org.codehaus.plexus.util.StringUtils;

import com.perforce.maven.scm.provider.p4.repository.P4ScmProviderRepository;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.client.IClientSummary;
import com.perforce.p4java.client.IClientViewMapping;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.FileSpecOpStatus;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.impl.generic.client.ClientView;
import com.perforce.p4java.impl.generic.client.ClientView.ClientViewMapping;
import com.perforce.p4java.impl.mapbased.client.Client;
import com.perforce.p4java.server.IServer;

/**
 * Manage IClient cache
 *
 * @plexus.component role="com.perforce.maven.scm.provider.p4.manager.P4ClientManager"
 */

public class DefaultP4ClientManager
    extends AbstractLogEnabled
    implements P4ClientManager, Disposable
{
    /**
     * @plexus.requirement role="com.perforce.maven.scm.provider.p4.manager.P4ServerManager"
     */
    private P4ServerManager serverManager;

    private Map<String, InternalClient> clients = new ConcurrentHashMap<String, InternalClient>();

    public void dispose()
    {
        for ( InternalClient internalClient : clients.values() )
        {
            if ( internalClient.getRemoveWhenDone() )
            {
                IClient client = internalClient.getClient();
                try
                {
                    client.getServer().deleteClient( client.getName(), false );
                    client.getServer().disconnect();
                }
                catch ( P4JavaException e )
                {
                    this.getLogger().error( e.getLocalizedMessage() );
                }
            }
        }

        clients.clear();
    }

    public IClient getClient( P4ScmProviderRepository repo, File rootDir )
        throws ScmException
    {
        return getClient( repo, rootDir, null );
    }

    public IClient getClient( P4ScmProviderRepository repo, File rootDir, String clientName )
        throws ScmException
    {

        if ( StringUtils.isBlank( clientName ) )
        {
            clientName = this.discoverClientName( repo, rootDir );
            if ( clientName == null )
            {
                clientName = generateClientName( repo, rootDir );
            }
        }

        IClient client = null;

        try
        {
            InternalClient internalClient = clients.get( getClientKey( repo, clientName, rootDir ) );

            if ( internalClient != null )
            {
                client = internalClient.getClient();
                if ( !isValidClient( client ) )
                {
                    this.getLogger().warn( "Staled cached client, creating new one ..." );
                    clients.remove( getClientKey( repo, clientName, rootDir ) );
                    client = null;
                }

            }

            if ( client == null )
            {
                boolean removeWhenDone = false;

                IServer server = serverManager.createServer( repo );

                client = server.getClient( clientName );
                if ( client == null )
                {
                    client = createNewClient( server, repo, clientName, rootDir );
                }

                if ( client.getName().equals( generateClientName( repo, rootDir ) ) )
                {
                    // this may be also what left off from previous process
                    removeWhenDone = true;
                }

                this.adjustClientSpecDueToP4JAVAIssue( client, repo );

                // no need to set working directory since it is very dynamic
                // each command must use absolute path instead
                // server.setWorkingDirectory( rootDir.getAbsolutePath() );

                this.clients.put( getClientKey( repo, clientName, rootDir ),
                                  new InternalClient( client, removeWhenDone ) );
            }

            client.getServer().setCurrentClient( client ); // so that we can sync
        }
        catch ( P4JavaException e )
        {
            throw new ScmException( e.getLocalizedMessage(), e );
        }
        catch ( UnknownHostException e )
        {
            throw new ScmException( e.getLocalizedMessage(), e );
        }

        return client;

    }

    private boolean isValidClient( IClient client )
    {
        return DefaultP4ServerManager.isValidServer( client.getServer() );
    }

    /**
     * adjust client view by additionally insert the intended entry from pom/connectionUrl if needed since the provide
     * client may not have it. There is also a bug which server incorrectly setup the client view after it is first
     * created, So we need to check and append a actual view to override the first one.
     *
     * @param client
     * @param repo
     * @throws P4JavaException
     */
    private void adjustClientSpecDueToP4JAVAIssue( IClient client, P4ScmProviderRepository repo )
        throws P4JavaException
    {
        ClientView clientView = client.getClientView();
        if ( clientView == null )
        {
            clientView = new ClientView();
        }

        ClientViewMapping clientViewMapping = createClientViewMapping( repo, client.getName() );
        List<IClientViewMapping> list = clientView.getEntryList();
        boolean exists = false;
        if ( list != null )
        {
            for ( IClientViewMapping map : list )
            {
                if ( map.getDepotSpec().equals( clientViewMapping.getDepotSpec() )
                    && map.getClient().equals( clientViewMapping.getClient() ) )
                {
                    exists = true;
                    break;
                }
            }
        }

        if ( !exists )
        {
            clientView.addEntry( clientViewMapping );
            client.setClientView( clientView );
            if ( client.canUpdate() )
            {
                client.update();
            }
            if ( client.canRefresh() )
            {
                client.refresh();
            }
        }

    }

    private IClient createNewClient( IServer server, P4ScmProviderRepository repo, String clientName, File basedir )
        throws P4JavaException, UnknownHostException
    {
        // Create a new client
        ClientViewMapping clientViewMapping = createClientViewMapping( repo, clientName );
        ClientView clientView = new ClientView();
        clientView.addEntry( clientViewMapping );

        IClient newClient =
            new Client( clientName, null, null, "Created by P4Maven", InetAddress.getLocalHost().getHostName(),
                        repo.getUser(), basedir.getAbsolutePath(), null, null, null, null, server, clientView );

        server.createClient( newClient );

        return server.getClient( clientName );

        // there is a bug, where the return IClient does not have the same requested ClientViewMapping
        // for example
        // //guest/dantran/p4maven/... //someclient/...
        // become
        // //guest/dantran/... //someclient/...
    }

    private String getClientKey( P4ScmProviderRepository repo, String clientName, File rootDir )
    {
        return repo.getP4Port() + ":" + clientName + ":" + rootDir.getAbsolutePath();
    }

    /**
     * Generate a Perforce client workspace name.
     *
     * @return the string
     */
    private String generateClientName( P4ScmProviderRepository repo, File basedir )
    {
        String username = repo.getUser();
        String hostname = repo.getHost();
        String path = null;
        try
        {
            hostname = InetAddress.getLocalHost().getHostName();
            path = encodeWildcards( basedir.getCanonicalPath().replaceAll( "[/ ~]", "-" ).replaceAll( ",", "" ) );
        }
        catch ( UnknownHostException e )
        {
            if ( getLogger().isErrorEnabled() )
            {
                this.getLogger().error( e.getMessage(), e );
            }
        }
        catch ( IOException e )
        {
            if ( getLogger().isErrorEnabled() )
            {
                this.getLogger().error( e.getMessage(), e );
            }
        }
        return username + "-" + hostname + "-MavenSCM-" + path + "-" + StringUtils.replace( repo.getPath(), '/', '\\' );
    }

    /**
     * Perforce wildcards expansion.
     *
     * @param filePath the file path
     * @return the string
     */
    public static String encodeWildcards( String filePath )
    {
        String path = new String();
        if ( filePath != null )
        {
            path =
                filePath.replaceAll( "%", "%25" ).replaceAll( "\\*", "%2A" ).replaceAll( "#", "%23" ).replaceAll( "@",
                                                                                                                  "%40" );
        }
        return path;
    }

    protected ClientViewMapping createClientViewMapping( P4ScmProviderRepository repo, String clientName )
    {
        // Create a new client
        String viewPath = getCanonicalRepoPath( repo.getPath() );
        return new ClientViewMapping( 0, viewPath, "//" + clientName + "/..." );
    }

    /**
     * Gets the canonical repo path.
     *
     * @param repoPath the repo path
     * @return the canonical repo path
     */
    public static String getCanonicalRepoPath( String repoPath )
    {
        if ( repoPath == null )
        {
            return null;
        }
        if ( repoPath.endsWith( "/..." ) )
        {
            return repoPath;
        }
        else if ( repoPath.endsWith( "/" ) )
        {
            return repoPath + "...";
        }
        else
        {
            return repoPath + "/...";
        }
    }

    private class InternalClient
    {
        private final IClient client;

        public IClient getClient()
        {
            return client;
        }

        public boolean getRemoveWhenDone()
        {
            return removeWhenDone;
        }

        private final boolean removeWhenDone;

        public InternalClient( IClient client, boolean removeWhenDone )
        {
            this.client = client;
            this.removeWhenDone = new Boolean( removeWhenDone );
        }
    }

    /**
     * From a given base directory ( ie a maven project with a pom.xml), look up and existing client that hosts this
     * project which may also be a sub project of the client's root directory.
     *
     * @param repo
     * @param baseDir
     * @return
     * @throws ScmException
     */
    public String discoverClientName( P4ScmProviderRepository repo, File baseDir )
        throws ScmException
    {
        IServer server = this.serverManager.getServer( repo );

        try
        {
            List<IClientSummary> clientInfos = server.getClients( repo.getUser(), null, Integer.MAX_VALUE );

            for ( IClientSummary clientInfo : clientInfos )
            {

                // make sure to doc this case, where the host field must match with current host
                if ( !InetAddress.getLocalHost().getHostName().equals( clientInfo.getHostName() ) )
                {
                    continue;
                }

                IClient client = server.getClient( clientInfo );

                if ( baseDir.equals( new File( clientInfo.getRoot() ) ) )
                {
                    // one dir may be used by 2 separate clients/depo paths, like maven-scm-tck branch test
                    if ( clientHasTheMatchingDepotPath( client, repo ) )
                    {
                        return clientInfo.getName();
                    }
                }
                else
                {
                    // see if repo's basedir is a subdirectory of an existing client
                    File clientRootDir = new File( clientInfo.getRoot() );
                    if ( clientRootDir.isDirectory() && FileUtils.directoryContains( clientRootDir, baseDir ) )
                    {
                        // see if the client really has this directory.
                        // IClient client = server.getClient( clientInfo );
                        // this.adjustClientSpecDueToP4JAVAIssue( client, repo );
                        server.setCurrentClient( client );

                        // assume this operation need a pom like release plugin
                        File currentPom = new File( baseDir, "pom.xml" );
                        if ( currentPom.exists() )
                        {
                            List<IFileSpec> remoteFiles =
                                server.getDepotFiles( client.where( FileSpecBuilder.makeFileSpecList( currentPom.getCanonicalPath() ) ),
                                                      false );
                            if ( remoteFiles.size() != 0 )
                            {
                                // check for error
                                IFileSpec fileSpec = remoteFiles.get( 0 );
                                if ( fileSpec.getOpStatus() == FileSpecOpStatus.VALID )
                                {
                                    return clientInfo.getName();
                                }
                            }
                        }
                    }
                }
            }
        }
        catch ( Exception e )
        {
            throw new ScmException( "Unable to discover client", e );
        }

        return null;
    }

    private boolean clientHasTheMatchingDepotPath( IClient client, P4ScmProviderRepository repo )
    {
        boolean found = false;
        for ( IClientViewMapping clientViewMapping : client.getClientView().getEntryList() )
        {
            if ( clientViewMapping.getLeft().startsWith( repo.getPath() + "/..." ) )
            {
                found = true;
                break;
            }
        }

        return found;
    }

}
